"""
Model utilities.
"""
import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras.models import Model

from hgd.config import Config


def make_model(
        seq_length: int = Config.seq_length,
        num_features: int = Config.num_features
) -> Model:
    seq_input = layers.Input(
        shape=(seq_length, num_features), dtype=tf.float32, name="input"
    )
    # Shape: (batch_size, seq_length, num_features)
    x = seq_input
    x = layers.Conv1D(2, 10, strides=5, padding="valid", activation="relu")(x)
    x = layers.Conv1D(1, 5, strides=2, padding="valid", activation="relu")(x)
    x_0 = layers.Flatten()(x)

    is_moving = layers.Dense(1, activation="sigmoid", name="is_moving")(x_0)
    class_prob = layers.Dense(len(Config.class_labels),
                              activation="softmax", name="class_prob")(x_0)
    output = layers.Concatenate()([is_moving, class_prob])
    model = Model(seq_input, output)
    return model



"""
Data utilities.
"""
import time
from typing import Dict, List, Sequence

import numpy as np
import numpy.typing

from hgd.config import Config
from hgd.video import video_to_landmarks

NDFloat32Array = np.typing.NDArray[np.float32]


def collect_landmarks_with_webcam(
        labels: Sequence[str] = (Config.stationary_label,) + Config.class_labels,
        max_num_frames: int = 800,
        output_npz_path: str = Config.npz_filename,
        sleep_seconds: float = 3.0
) -> None:
    landmark_dict: Dict[str, List[List[float]]] = {}
    for label in labels:
        landmark_dict[label] = video_to_landmarks(None, max_num_frames)
        time.sleep(sleep_seconds)
    np.savez_compressed(output_npz_path, **landmark_dict)


def preprocess(landmarks: List[List[float]]) -> NDFloat32Array:
    features = np.array(landmarks)

    # Scale with face box coordinates.
    features[:, 4:10] = (features[:, 4:10] - features[:, 0:1]) / features[:, 2:3]
    features[:, 10:] = (features[:, 10:] - features[:, 1:2]) / features[:, 3:4]

    # Add stddev features.
    xs_coord_std = np.std(features[:, 4:10], axis=1, keepdims=True)
    ys_coord_std = np.std(features[:, 10:], axis=1, keepdims=True)
    features = np.concatenate([features, xs_coord_std, ys_coord_std], axis=1)

    # Drop face boxes.
    features = np.delete(features, [0, 1, 2, 3], axis=1)

    assert features.shape[1] == Config.num_features
    return features


if __name__ == "__main__":
    collect_landmarks_with_webcam()



"""
Config.
"""
from dataclasses import dataclass
from typing import Tuple


@dataclass
class Config:
    class_labels: Tuple[str] = ("head_nodding", "head_turning")
    stationary_label: str = "stationary"
    npz_filename: str = "landmarks.npz"
    seq_length: int = 60
    num_features: int = 14
    weights_filename: str = "hgd_weights.h5"



"""
Inference utilities.
"""
import os
from typing import Callable, List, Optional, Sequence

import numpy as np
from tensorflow.keras.models import Model

from hgd._download import get_default_weights_path, download_weights_to
from hgd.config import Config
from hgd.data import NDFloat32Array, preprocess
from hgd.model import make_model
from hgd.video import video_to_landmarks


def load_pretrained_model(weights_path: str = get_default_weights_path()) -> Model:
    model = make_model()
    if not os.path.isfile(weights_path):
        download_weights_to(weights_path)
    model.load_weights(weights_path)
    return model


def predict_video(
        video_path: Optional[str] = None,  # None will start a webcam.
        model: Optional[Model] = None,
        max_num_frames: int = Config.seq_length,  # For the pre-trained model.
        padding: bool = True,
        preprocess_fn: Callable[[List[List[float]]], NDFloat32Array] = preprocess
) -> Sequence[float]:
    if model is None:
        model = load_pretrained_model()
    landmarks = video_to_landmarks(video_path, max_num_frames, padding)
    prediction: Sequence[float] = model.predict(
        np.expand_dims(preprocess_fn(landmarks), axis=0)
    )[0].tolist()
    return prediction


# if __name__ == "__main__":
#     print(predict_video(model=load_pretrained_model(f"../training/{Config.weights_filename}")))
